import os
import subprocess
import sys

from celescope.tools import utils
from celescope.tools.make_ref import MakeRef, GENOME_CONFIG


class Mkref_snp:
    def __init__(self):
        pass

    @utils.add_log
    def build_fasta_index(self, fasta):
        cmd = f"samtools faidx {fasta}"
        sys.stderr.write(cmd + "\n")
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def build_fasta_dict(self, fasta):
        cmd = f"gatk CreateSequenceDictionary -R {fasta}"
        sys.stderr.write(cmd + "\n")
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def run(self):
        assert os.path.exists(GENOME_CONFIG), (
            f"{GENOME_CONFIG} not exist! "
            "You need to build rna genome with 'celescope rna mkref' first."
        )
        # reset genome type here
        config = MakeRef.get_config("./")
        fasta = config["files"]["fasta"]
        self.build_fasta_index(fasta)
        self.build_fasta_dict(fasta)

        config["meta"]["genome_type"] = "rna, snp"
        with open(GENOME_CONFIG, "w") as config_handle:
            config.write(config_handle)


def mkref(args):
    # pylint: disable=unused-argument
    Mkref_snp().run()


def get_opts_mkref(parser, sub_program):
    # pylint: disable=unused-argument
    pass



from celescope.snp.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_snp(Multi):
    """
    ## Usage

    ### Make a snp reference genomeDir

    1. Run `celescope rna mkref`. If you already have a rna genomeDir, you can use it and skip this step.
    2. Run `celescope snp mkref` under the rna genomeDir. Check [mkref.md](./mkref.md) for help.

    ### Run multi_snp
    There are two ways to run `multi_snp`

    1. Do not perform consensus before alignment and report read count(recommended for data generated with FocuSCOPE kit).

    ```
    multi_snp\\
        --mapfile ./test1.mapfile\\
        --genomeDir {genomeDir after running celescope snp mkref}\\
        --thread 4\\
        --mod shell\\
        --panel lung_1\\
        --not_consensus
    ```

    2. Do consensus before alignment and report UMI count. 

    ```
    multi_snp\\
        --mapfile ./test1.mapfile\\
        --genomeDir {genomeDir after running celescope snp mkref}\\
        --thread 4\\
        --mod shell\\
        --panel lung_1\\
    ```

    """

    def target_metrics(self, sample):
        step = "target_metrics"
        cmd_line = self.get_cmd_line(step, sample)
        bam = f'{self.outdir_dic[sample]["starsolo"]}/{sample}_Aligned.sortedByCoord.out.bam'
        cmd = (
            f"{cmd_line} "
            f"--bam {bam} "
            f"--match_dir {self.col4_dict[sample]} "
            f"--add_RG "
        )
        self.process_cmd(cmd, step, sample, m=2, x=1)

    def variant_calling(self, sample):
        step = "variant_calling"
        cmd_line = self.get_cmd_line(step, sample)
        bam = (
            f'{self.outdir_dic[sample]["target_metrics"]}/{sample}_filtered_sorted.bam'
        )
        cmd = f"{cmd_line} " f"--bam {bam} " f"--match_dir {self.col4_dict[sample]} "
        self.process_cmd(cmd, step, sample, m=8, x=1)

    def filter_snp(self, sample):
        step = "filter_snp"
        vcf = f'{self.outdir_dic[sample]["variant_calling"]}/{sample}_norm.vcf'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = f"{cmd_line} " f"--vcf {vcf} "
        self.process_cmd(cmd, step, sample, m=1, x=1)

    def analysis_snp(self, sample):
        step = "analysis_snp"
        vcf = f'{self.outdir_dic[sample]["filter_snp"]}/{sample}_filtered.vcf'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = f"{cmd_line} " f"--match_dir {self.col4_dict[sample]} " f"--vcf {vcf} "
        self.process_cmd(cmd, step, sample, m=2, x=1)


def main():
    multi = Multi_snp(__ASSAY__, min_col=4)
    multi.run()


if __name__ == "__main__":
    main()



"""keep the same amount of reads in each well barcode"""

import argparse
from collections import defaultdict

import pysam

import celescope.tools.parse_chemistry as parse_chemistry
from celescope.__init__ import HELP_DICT
from celescope.chemistry_dict import chemistry_dict
from celescope.tools import utils


def main():
    parser = get_parser(argparse.ArgumentParser())
    args = parser.parse_args()

    fq1_list = args.fq1.split(",")
    fq2_list = args.fq2.split(",")

    chemistry = parse_chemistry.get_chemistry("bulk_rna", args.chemistry, fq1_list)
    pattern_dict, bc = parse_chemistry.get_pattern_dict_and_bc(
        chemistry, args.pattern, args.whitelist
    )
    raw_list, mismatch_list = (
        parse_chemistry.create_mismatch_origin_dicts_from_whitelists(bc, 1)
    )

    barcode_readcount = defaultdict(int)
    out_r1_fn = f"{args.outdir}/out_R1.fq.gz"
    out_r2_fn = f"{args.outdir}/out_R2.fq.gz"
    out_r1 = utils.generic_open(out_r1_fn, "wt", compresslevel=1)
    out_r2 = utils.generic_open(out_r2_fn, "wt", compresslevel=1)

    for fq1, fq2 in zip(fq1_list, fq2_list):
        fq1 = pysam.FastxFile(fq1)
        fq2 = pysam.FastxFile(fq2)
        for e1, e2 in zip(fq1, fq2):
            bc_list = [e1.sequence[x] for x in pattern_dict["C"]]
            valid, corrected, corrected_bc = parse_chemistry.check_seq_mismatch(
                bc_list, raw_list, mismatch_list
            )
            if valid:
                barcode_readcount[corrected_bc] += 1
                if barcode_readcount[corrected_bc] <= args.max_read:
                    out_r1.write(utils.fastq_line(e1.name, e1.sequence, e1.quality))
                    out_r2.write(utils.fastq_line(e2.name, e2.sequence, e2.quality))

    out_r1.close()
    out_r2.close()


def get_parser(parser):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--chemistry",
        help=HELP_DICT["chemistry"],
        choices=list(chemistry_dict.keys()),
        default="auto",
    )
    parser.add_argument(
        "--pattern",
        help="""The pattern of R1 reads, e.g. `C8L16C8L16C8L1U12T18`. The number after the letter represents the number 
        of bases.  
        - `C`: cell barcode  
        - `L`: linker(common sequences)  
        - `U`: UMI    
        - `T`: poly T""",
    )
    parser.add_argument(
        "--whitelist",
        help="Cell barcode whitelist file path, one cell barcode per line.",
    )
    parser.add_argument(
        "--fq1",
        help="R1 fastq file. Multiple files are separated by comma.",
        required=True,
    )
    parser.add_argument(
        "--fq2",
        help="R2 fastq file. Multiple files are separated by comma.",
        required=True,
    )
    parser.add_argument(
        "--max_read",
        help="keep maximum this number of reads per well in the output fastq files",
        required=True,
        type=int,
    )
    parser.add_argument(
        "--outdir",
        help="Output directory",
        default="./",
    )
    return parser


if __name__ == "__main__":
    main()



import itertools
import os
import re
import sys
from collections import defaultdict

import pysam
from celescope.tools import utils
from celescope.chemistry_dict import chemistry_dict, chemistry_dir


def parse_pattern(pattern: str, allowed: str = "CLUNT") -> dict[str, list[slice]]:
    """Parse a pattern string into a dictionary of slices.

    >>> pattern_dict = parse_pattern("C8L16C8L16C8L1U12T18")
    >>> pattern_dict['C']
    [slice(0, 8, None), slice(24, 32, None), slice(48, 56, None)]
    >>> pattern_dict['L']
    [slice(8, 24, None), slice(32, 48, None), slice(56, 57, None)]
    """
    pattern_slices: dict[str, list[slice]] = {}
    p = re.compile(r"([A-Z])(\d+)")  # Compile the regex

    start = 0
    for char, length in p.findall(pattern):
        if char not in allowed:
            raise ValueError(f"Invalid character '{char}' in pattern: {pattern}")
        if char not in pattern_slices:
            pattern_slices[char] = []
        end = start + int(length)
        pattern_slices[char].append(slice(start, end))
        start = end
    return pattern_slices


def create_mismatch_seqs(seq: str, max_mismatch=1, allowed_bases="ACGTN") -> set[str]:
    """Create all sequences within a specified number of mismatches from the input sequence.

    >>> answer = set(["TCG", "AAG", "ACC", "ATG", "ACT", "ACN", "GCG", "ANG", "ACA", "ACG", "CCG", "AGG", "NCG"])
    >>> seq_set = create_mismatch_seqs("ACG")
    >>> seq_set == answer
    True
    >>> seq_set = create_mismatch_seqs("ACG", max_mismatch=0)
    >>> seq_set == set(["ACG"])
    True
    """
    if max_mismatch < 0:
        raise ValueError("max_mismatch must be non-negative")
    if max_mismatch > len(seq):
        raise ValueError(
            f"max_mismatch ({max_mismatch}) cannot be greater than the sequence length ({len(seq)})"
        )

    result = set()
    for n_mismatch in range(max_mismatch + 1):  # 包括0到max_mismatch
        for locs in itertools.combinations(range(len(seq)), n_mismatch):
            seq_locs = [
                list(allowed_bases) if i in locs else [base]
                for i, base in enumerate(seq)
            ]
            result.update("".join(p) for p in itertools.product(*seq_locs))
    return result


def create_mismatch_origin_dict(origin_seqs: list, n_mismatch=1) -> dict[str, str]:
    """Create a dictionary mapping sequences with mismatches to their original sequences(in whitelist).

    >>> origin_seqs = ["AACGTGAT", "AAACATCG"]
    >>> mismatch_dict = create_mismatch_origin_dict(origin_seqs)
    >>> mismatch_dict["AACGTGAA"] == "AACGTGAT"
    True
    """
    result = {}
    for origin_seq in origin_seqs:
        origin_seq = origin_seq.strip()
        if origin_seq == "":
            continue
        for mismatch_seq in create_mismatch_seqs(origin_seq, n_mismatch):
            result[mismatch_seq] = origin_seq
    return result


def create_mismatch_origin_dicts_from_whitelists(
    whitelists: list, n_mismatch: int = 1
) -> tuple[list, list]:
    """Returns raw dict list and mismatch dict list.

    >>> whitelists = [os.path.join(chemistry_dir, "GEXSCOPE-V1/bc.txt")]
    >>> raw_list, mismatch_list = create_mismatch_origin_dicts_from_whitelists(whitelists)
    >>> len(raw_list) == len(mismatch_list)
    True
    """
    raw_list, mismatch_list = [], []
    for f in whitelists:
        barcodes = utils.one_col_to_list(f)
        raw_list.append(set(barcodes))
        barcode_mismatch_dict = create_mismatch_origin_dict(barcodes, n_mismatch)
        mismatch_list.append(barcode_mismatch_dict)

    return raw_list, mismatch_list


def check_seq_mismatch(seq_list, raw_list, mismatch_list):
    """
    Returns
        valid: True if seq in mismatch_list or mismatch_list is empty
        corrected: True if seq in mismatch_list but not in raw_list
        res: joined seq

    >>> seq_list = ['ATA', 'AAT', 'ATA']
    >>> correct_set_list = [{'AAA'},{'AAA'},{'AAA'}]
    >>> mismatch_dict_list = [create_mismatch_origin_dict(['AAA'])] * 3

    >>> check_seq_mismatch(seq_list, correct_set_list, mismatch_dict_list)
    (True, True, 'AAA_AAA_AAA')

    >>> seq_list = ['AAA', 'AAA', 'AAA']
    >>> check_seq_mismatch(seq_list, correct_set_list, mismatch_dict_list)
    (True, False, 'AAA_AAA_AAA')

    >>> seq_list = ['AAA', 'AAA', 'AAA']
    >>> raw_list, mismatch_list = [], []
    >>> check_seq_mismatch(seq_list, raw_list, mismatch_list)
    (True, False, 'AAA_AAA_AAA')
    """
    if not mismatch_list:
        return True, False, "_".join(seq_list)
    valid = True
    corrected = False
    res = []
    for index, seq in enumerate(seq_list):
        if seq not in raw_list[index]:
            if seq not in mismatch_list[index]:
                valid = False
                res.append("")
            else:
                corrected = True
                res.append(mismatch_list[index][seq])
        else:
            res.append(seq)

    return valid, corrected, "_".join(res)


def get_chemistry_dict():
    """
    Return:
    chemistry_dict. Key: chemistry name, value: chemistry dict

    >>> chemistry_dict = get_chemistry_dict()
    >>> chemistry_dict["GEXSCOPE-MicroBead"]["pattern_dict"]
    {'C': [slice(0, 12, None)], 'U': [slice(12, 20, None)]}
    """
    # add folder prefix
    for chemistry in chemistry_dict:
        cur = chemistry_dict[chemistry]
        bc = cur.get("bc", [])
        linker = cur.get("linker", [])
        if bc:
            cur["bc"] = [os.path.join(chemistry_dir, chemistry, x) for x in bc]
        if linker:
            cur["linker"] = [os.path.join(chemistry_dir, chemistry, x) for x in linker]
        cur["pattern_dict"] = parse_pattern(cur["pattern"])
    return chemistry_dict


CHEMISTRY_DICT = get_chemistry_dict()


def get_raw_umi_bc_and_quality(
    seq: str, quality: str, pattern_dict: dict, reverse_complement=False
) -> tuple[list, list, str, str]:
    """
    Returns:
        bc_list, umi, bc_quality_list, umi_qual
    """
    bc_list = [seq[x] for x in pattern_dict["C"]]
    bc_quality_list = [quality[x] for x in pattern_dict["C"]]
    umi = seq[pattern_dict["U"][0]]
    umi_qual = quality[pattern_dict["U"][0]]
    if reverse_complement:
        bc_list = [utils.reverse_complement(x) for x in bc_list[::-1]]
        bc_quality_list = bc_quality_list[::-1]
        umi = utils.reverse_complement(umi)
        umi_qual = umi_qual[::-1]
    return bc_list, bc_quality_list, umi, umi_qual


class Auto:
    """
    Auto detect singleron chemistrys from R1-read
    """

    def __init__(self, fq1_list, chemistry_dict, max_read=10000):
        """
        Returns:
            chemistry, chemistry_dict[chemistry]
        """
        self.fq1_list = fq1_list
        self.max_read = max_read
        self.chemistry_dict = chemistry_dict
        self.mismatch_dict = {}
        for chemistry in self.chemistry_dict:
            if "bc" in self.chemistry_dict[chemistry]:
                self.mismatch_dict[chemistry] = (
                    create_mismatch_origin_dicts_from_whitelists(
                        self.chemistry_dict[chemistry]["bc"], 1
                    )
                )

    def run(self):
        """
        Returns:
            chemistry, chemistry_dict[chemistry]
        """
        chemistry = self.get_chemistry()
        return chemistry, self.chemistry_dict[chemistry]

    def get_chemistry(self) -> str:
        """check chemistry in the fq1_list"""
        fq_chemistry = {}
        for fastq1 in self.fq1_list:
            chemistry = self.get_fq_chemistry(fastq1)
            fq_chemistry[fastq1] = chemistry
        if len(set(fq_chemistry.values())) != 1:
            sys.exit(
                f"Error: multiple chemistrys are not allowed for one sample: {self.fq1_list}! \n"
                + str(fq_chemistry)
            )
        chemistry = list(fq_chemistry.values())[0]
        return chemistry

    def is_chemistry(self, seq, chemistry):
        """check if seq matches the barcode of chemistry"""
        raw_list, mismatch_list = self.mismatch_dict[chemistry]
        bc_list = [seq[x] for x in self.chemistry_dict[chemistry]["pattern_dict"]["C"]]
        if chemistry.split("-")[0] == "flv":
            bc_list = [utils.reverse_complement(bc) for bc in bc_list[::-1]]
        valid, _corrected, _res = check_seq_mismatch(bc_list, raw_list, mismatch_list)
        return valid

    def seq_chemistry(self, seq):
        """check if seq matches any chemistry in chemistry_dict"""
        for chemistry in self.chemistry_dict:
            if self.is_chemistry(seq, chemistry):
                return chemistry
        return None

    def get_fq_chemistry(self, fq1):
        chemistry_readcount = defaultdict(int)

        fq = pysam.FastxFile(fq1)
        n = 0
        for read in fq:
            seq = read.sequence
            n += 1
            chemistry = self.seq_chemistry(seq)
            if chemistry:
                chemistry_readcount[chemistry] += 1
            if n == self.max_read:
                break
        sorted_counts = sorted(
            chemistry_readcount.items(), key=lambda x: x[1], reverse=True
        )
        print(sorted_counts)

        chemistry, read_counts = sorted_counts[0]
        percent = float(read_counts) / n
        if percent < 0.1:
            print("Valid chemistry read counts percent < 0.1")
            raise Exception("Auto chemistry detection failed! ")
        elif percent < 0.5:
            print("Valid chemistry read counts percent < 0.5")
        print(f"{fq1}: {chemistry}")

        return chemistry


FLV_RNA_V2_LINKER1 = "ATCCAGCTGCTTGAGATC"


class AutoRNA(Auto):
    def __init__(self, fq1_list, max_read=10000):
        super().__init__(fq1_list, CHEMISTRY_DICT, max_read)
        self.v3_linker_mismatch = create_mismatch_origin_dicts_from_whitelists(
            self.chemistry_dict["GEXSCOPE-V3"]["linker"], 1
        )
        self.flv_rna_v2_linker1_mismatch_dict = create_mismatch_origin_dict(
            [FLV_RNA_V2_LINKER1], 1
        )

    def v3_offset(self, seq):
        """
        return -1 if not v3

        >>> seq = "AT" + "TCGACTGTC" + "ACGATG" + "TTCTAGGAT" + "CATAGT" + "TGCACGAGA" + "C" + "CATATCAATGGG" + "TTTTTTTTTT"
        >>> runner = AutoRNA([], "fake_sample")
        >>> runner.v3_offset(seq)
        2
        >>> seq = "TCGACTGTC" + "ACGATG" + "TTCTAGGAT" + "CATAGT" + "TGCACGAGA" + "C" + "CATATCAATGGG" + "TTTTTTTTTT"
        >>> runner.v3_offset(seq)
        0
        >>> seq = "TCGACTGTC" + "ATATAT" + "TTCTAGGAT" + "CATAGT" + "TGCACGAGA" + "C" + "CATATCAATGGG" + "TTTTTTTTTT"
        >>> runner.v3_offset(seq)
        -1
        """
        bc_len = 9
        linker_len = 6
        max_offset_len = 3 + 1  # allow for extra 1 bases
        for offset in range(max_offset_len + 1):
            first_linker_start = offset + bc_len
            second_linker_start = first_linker_start + linker_len + bc_len
            first_linker_seq = seq[first_linker_start : first_linker_start + linker_len]
            second_linker_seq = seq[
                second_linker_start : second_linker_start + linker_len
            ]
            valid, _, _ = check_seq_mismatch(
                [first_linker_seq, second_linker_seq], *self.v3_linker_mismatch
            )
            if valid:
                return offset
        return -1

    def flv_rna_v2_offset(self, seq) -> int:
        """
        return -1 if not

        >>> seq = FLV_RNA_V2_LINKER1
        >>> runner = AutoRNA([], "fake_sample")
        >>> runner.flv_rna_v2_offset(seq)
        0
        >>> seq = "A" + FLV_RNA_V2_LINKER1
        >>> runner.flv_rna_v2_offset(seq)
        1
        >>> seq = "GGGGG" + FLV_RNA_V2_LINKER1
        >>> runner.flv_rna_v2_offset(seq)
        -1
        """
        max_offset_len = 3 + 1  # allow for extra 1 bases
        for offset in range(max_offset_len + 1):
            linker = seq[offset : offset + 18]
            if linker in self.flv_rna_v2_linker1_mismatch_dict:
                return offset
        return -1

    def seq_chemistry(self, seq):
        """
        Returns: chemistry or None

        >>> import tempfile
        >>> runner = AutoRNA([], "fake_sample")
        >>> seq = "AT" + "TCGACTGTC" + "ACGATG" + "TTCTAGGAT" + "CATAGT" + "TGCACGAGA" + "C" + "CATATCAATGGG" + "TTTTTTTTTT"
        >>> runner.seq_chemistry(seq)
        'GEXSCOPE-V3'
        >>> seq = "TCGACTGTC" + "ATCCACGTGCTTGAGA" + "TTCTAGGAT" + "TCAGCATGCGGCTACG" + "TGCACGAGA" + "C" + "CATATCAATGGG" + "TTTTTTTTTT"
        >>> runner.seq_chemistry(seq)
        'GEXSCOPE-V2'
        >>> seq = "NCAGATTC" + "TCGGTGACAGCCATAT" + "GTACGCAA" + "CGTAGTCAGAAGCTGA" + "CTGAGCCA" + "C" + "TCCGAAGCCCAT" + "TTTTTTTTTT"
        >>> runner.seq_chemistry(seq)
        'GEXSCOPE-V1'
        >>> seq = "NCAGATTC" + "TCGGTGACAGCCATAT" + "GTACGCAA" + "CGTAGTCAGAAGCTGA" + "CTGAGCCA"  + "TCCGAAGCC" + "CTGTCT"
        >>> runner.seq_chemistry(seq)
        'flv_rna'
        >>> seq = "NCAGATTC" + "TCGGTGACAGCCATAT" + "GTACGCAA" + "CGTAGTCAGAAGCTGA" + "CTGAGCCA"  + "C" + "TCCGAAGCC"
        >>> runner.seq_chemistry(seq)
        'GEXSCOPE-V1'
        >>> seq = "ATCGATCGATCG" + "ATCGATCG" + "C" + "TTTTTTTTTT"
        >>> runner.seq_chemistry(seq)
        'GEXSCOPE-MicroBead'
        """
        if self.v3_offset(seq) != -1:
            return "GEXSCOPE-V3"

        if self.flv_rna_v2_offset(seq) != -1:
            return "flv_rna-V2"

        for chemistry in ["GEXSCOPE-V2", "GEXSCOPE-V1"]:
            if self.is_chemistry(seq, chemistry):
                if chemistry == "GEXSCOPE-V1":
                    if seq[56] != "C":
                        return "flv_rna"
                return chemistry

        # check if it is MicroBead
        if seq[16:20] != "TTTT" and seq[22:26] == "TTTT":
            return "GEXSCOPE-MicroBead"


class AutoBulkRNA(Auto):
    def __init__(self, fq1_list, max_read=10000):
        super().__init__(fq1_list, CHEMISTRY_DICT, max_read)

    def seq_chemistry(self, seq):
        """
        Returns: chemistry or None
        """
        if seq[:18] == "GTGGTATCAACGCAGAGT":
            return "bulk_rna-bulk_vdj_match"
        # V2 9bp linker is ATACGCGGA, which is a valid barcode of V1; so must detect V2 first, otherwise it is a valid V1
        for chemistry in ["bulk_rna-V2", "bulk_rna-V1", "bulk_rna-V3"]:
            if self.is_chemistry(seq, chemistry):
                return chemistry


class AutoSpace(Auto):
    def __init__(self, fq1_list, max_read=10000):
        super().__init__(fq1_list, CHEMISTRY_DICT, max_read)

    def seq_chemistry(self, seq):
        """
        Returns: chemistry or None
        """
        for chemistry in ["space-ffpe", "space-ff"]:
            if self.is_chemistry(seq, chemistry):
                return chemistry


class AutoFlv(Auto):
    def __init__(self, fq1_list, max_read=10000):
        super().__init__(fq1_list, CHEMISTRY_DICT, max_read)

    def seq_chemistry(self, seq):
        """
        Returns: chemistry or None
        """
        for chemistry in ["flv", "flv-V2"]:
            if self.is_chemistry(seq, chemistry):
                return chemistry


@utils.add_log
def get_chemistry(assay: str, args_chemistry: str, fq1_list: list) -> str:
    """Auto detect chemistry. If customized, return 'customized'"""
    if args_chemistry == "auto":
        if assay in ["bulk_vdj"]:
            return assay
        elif assay == "space":
            return AutoSpace(fq1_list).get_chemistry()
        elif assay == "flv_trust4":
            return AutoFlv(fq1_list).get_chemistry()
        elif assay == "bulk_rna":
            return AutoBulkRNA(fq1_list).get_chemistry()
        else:
            return AutoRNA(fq1_list).get_chemistry()
    else:
        return args_chemistry


@utils.add_log
def get_pattern_dict_and_bc(
    chemistry, pattern: str = "", whitelist: str = ""
) -> tuple[dict, list]:
    if chemistry != "customized":
        chemistry_dict = get_chemistry_dict()
        pattern_dict = chemistry_dict[chemistry]["pattern_dict"]
        bc = chemistry_dict[chemistry]["bc"]
    else:
        pattern_dict = parse_pattern(pattern)
        bc = whitelist.split(" ")
    return pattern_dict, bc


def invalid_debug(
    chemistry,
    fq1,
    output_file,
    use_read=1000,
    skip_read=10000,
    pattern: str = None,
    whitelist: str = None,
    linker: str = None,
):
    cur = CHEMISTRY_DICT[chemistry]
    if chemistry == "customized":
        cur["bc"] = whitelist.split(" ") if whitelist else []
        cur["pattern"] = pattern
        cur["linker"] = linker.split(" ") if linker else []
    fq = pysam.FastxFile(fq1)
    bcs, linkers = [], []
    if "bc" in cur:
        for x in cur["bc"]:
            bc = utils.one_col_to_list(x)
            if chemistry.split("-")[0] == "flv":
                bc = [utils.reverse_complement(x) for x in bc]
            bcs.append(bc)
    if "linker" in cur:
        for x in cur["linker"]:
            linker = utils.one_col_to_list(x)
            if chemistry.split("-")[0] == "flv":
                linker = [utils.reverse_complement(x) for x in linker]
            linkers.append(linker)
    runner = BcUmi(chemistry, pattern, whitelist)
    html_sequences = []
    n_invalid = 0
    n_read = 0
    for read in fq:
        n_read += 1
        if n_read < skip_read:
            continue
        if n_read >= skip_read + use_read:
            break
        seq = read.sequence
        valid, _corrected, _corrected_seq, _umi = runner.get_bc_umi(seq)
        if valid:
            continue
        n_invalid += 1
        for bc, color in zip(bcs, ["red", "green", "blue"]):
            seq = add_color_in_html(seq, bc, background_color=color)
        for linker, color in zip(linkers, ["yellow", "pink", "orange"]):
            seq = add_color_in_html(seq, linker, background_color=color)
        html_sequences.append(f"{n_invalid}--{read.name}")
        html_sequences.append(seq)

    joined_sequences = "<br>".join(html_sequences)

    html_content = f"""
    <!DOCTYPE html>
    <html>
    <head>
        <meta charset="UTF-8">
    </head>
    <body>
        <h3>Color Legend</h3>
            <p>
                bc1: <span style="background-color:red;">red</span>
                bc2: <span style="background-color:green;">green</span>
                bc3: <span style="background-color:blue;">blue</span><br>
                linker1: <span style="background-color:yellow;">yellow</span>
                linker2: <span style="background-color:pink;">pink</span>
                linker3: <span style="background-color:orange;">orange</span>
            </p>
        <h3> invalid reads in number {skip_read+1} to {skip_read + use_read} reads</h3>
        {joined_sequences}
    </body>
    </html>
    """

    with open(output_file, "w", encoding="utf-8") as f:
        f.write(html_content)
    sys.stderr.write(f"invalid reads in {use_read} reads:{n_invalid}\n")


def add_color_in_html(seq, items, color="black", background_color="white"):
    for item in items:
        seq = seq.replace(
            item,
            f'<span style="color:{color};background-color:{background_color};">{item}</span>',
        )
    return seq


class BcUmi:
    def __init__(self, chemistry, pattern=None, whitelist=None):
        self.chemistry = chemistry
        self.pattern_dict, self.bc = get_pattern_dict_and_bc(
            self.chemistry, pattern, whitelist
        )
        self.raw_list, self.mismatch_list = (
            create_mismatch_origin_dicts_from_whitelists(self.bc, 1)
        )
        # v3
        self.offset_runner = AutoRNA([])

    def get_bc_umi(self, seq):
        if self.chemistry == "GEXSCOPE-V3":
            offset = self.offset_runner.v3_offset(seq)
            seq = seq[offset:]
        elif self.chemistry == "flv_rna-V2":
            offset = self.offset_runner.flv_rna_v2_offset(seq)
            seq = seq[offset:]
        bc_list = [seq[x] for x in self.pattern_dict["C"]]
        if self.chemistry.split("-")[0] == "flv":
            bc_list = [utils.reverse_complement(bc) for bc in bc_list[::-1]]
        valid, corrected, corrected_seq = check_seq_mismatch(
            bc_list, self.raw_list, self.mismatch_list
        )
        if not valid:
            umi = None
        else:
            umi = seq[self.pattern_dict["U"][0]]
        return valid, corrected, corrected_seq, umi



import shutil
import sys
import h5py
import numpy as np
import pandas as pd
from pathlib import Path
import scanpy as sc
import json
from PIL import Image

from celescope.tools import utils


def resize_by_longest_side(img, target_longest):
    """将图片最长边缩放到 target_longest 像素"""
    w, h = img.size
    if w >= h:
        # 横图：宽是最长边
        scale = target_longest / w
    else:
        # 竖图：高是最长边
        scale = target_longest / h
    new_size = (int(w * scale), int(h * scale))
    return img.resize(new_size, Image.LANCZOS)


class Spatial:
    """
    take a Abrowser spatial as input and output a visium like spatial directory
    """

    def __init__(self, input_dir):
        self.input_dir = Path(input_dir)
        self.tissue_positions_list = self.input_dir / "tissue_positions_list.csv"
        self.scalefactors = self.input_dir / "scalefactors_json.json"
        self.tissue_loweres = self.input_dir / "tissue_lowres_image.png"
        self.tissue_hires = self.input_dir / "tissue_hires_image.png"
        self.parquet = self.input_dir / "tissue_positions.parquet"
        figure_dir = self.input_dir / "figure"
        self.he_image = (
            list(figure_dir.glob("*.png"))
            + list(figure_dir.glob("*.jpg"))
            + list(figure_dir.glob("*.jpeg"))
        )
        if self.he_image:
            self.he_image = self.he_image[0]
        else:
            sys.exit("No HE .png, .jpg, .jpeg found in figure folder!")

        self.positions = pd.read_csv(self.tissue_positions_list, header=None)
        self.positions.columns = [
            "barcode",
            "in_tissue",
            "array_row",
            "array_col",
            "pxl_row_in_fullres",
            "pxl_col_in_fullres",
        ]

    def get_in_tissue_barcodes(self) -> list[str]:
        barcodes_in_tissue = self.positions.loc[
            self.positions["in_tissue"] == 1, "barcode"
        ].tolist()
        return barcodes_in_tissue

    @utils.add_log
    def output_spatial(self, outdir):
        outdir = Path(outdir)
        shutil.copytree(self.input_dir, outdir, dirs_exist_ok=True)
        parquet = outdir / "tissue_positions.parquet"
        self.positions.to_parquet(parquet, index=False)

        # add "microns_per_pixel"，"regist_target_img_scalef" to scalefactors
        scalefactors = outdir / "scalefactors_json.json"
        with open(scalefactors) as f:
            data = json.load(f)
        data["microns_per_pixel"] = 1.0
        data["regist_target_img_scalef"] = 1.0
        with open(scalefactors, "w") as f:
            json.dump(data, f, indent=4)

        ## output 2000 pixel HE hires and 600 pixel lowres
        Image.MAX_IMAGE_PIXELS = None
        he_image = Image.open(self.he_image)
        img_2000 = resize_by_longest_side(he_image, 2000)
        hires_path = outdir / "tissue_hires_image.png"
        img_2000.save(hires_path)

        img_600 = resize_by_longest_side(he_image, 600)
        lowres_path = outdir / "tissue_lowres_image.png"
        img_600.save(lowres_path)

    def rename_tissue_positions_csv(self, outdir):
        outdir = Path(outdir)
        csv_file = outdir / "tissue_positions_list.csv"
        if csv_file.exists():
            csv_file.rename(outdir / "positions_list.csv")


def convert_10x_h5(mtx_dir, outfile, library_id="library0"):
    """
    Convert a 10X MTX folder to a Visium-compatible HDF5 file.
    Existing file will be overwritten.

    Args:
        mtx_dir (str or Path): Path to 10X-formatted mtx directory.
        outfile (str or Path): Output HDF5 file path.
        library_id (str): Library ID for Visium metadata.
    """
    mtx_dir = Path(mtx_dir)
    outfile = Path(outfile)
    if not outfile.suffix == ".h5":
        outfile = outfile.with_suffix(".h5")

    # 如果文件已存在，直接覆盖
    if outfile.exists():
        outfile.unlink()

    # 读取 MTX 文件
    adata = sc.read_10x_mtx(mtx_dir, var_names="gene_symbols", make_unique=True)

    n_cells = adata.n_obs
    n_genes = adata.n_vars

    # 默认 feature_type 和 genome
    feature_types = (
        adata.var["feature_types"]
        if "feature_types" in adata.var.columns
        else ["Gene Expression"] * n_genes
    )
    genomes = (
        adata.var["genome"] if "genome" in adata.var.columns else ["unknown"] * n_genes
    )
    gene_ids = (
        adata.var["gene_ids"] if "gene_ids" in adata.var.columns else adata.var_names
    )

    # 创建 HDF5 文件
    with h5py.File(outfile, "w") as f:
        grp = f.create_group("matrix")

        # barcodes
        grp.create_dataset(
            "barcodes",
            data=np.array(adata.obs_names, dtype=f"|S{max(map(len, adata.obs_names))}"),
        )

        # counts
        grp.create_dataset("data", data=adata.X.data.astype(np.int32))
        grp.create_dataset("indices", data=adata.X.indices.astype(np.int32))
        grp.create_dataset("indptr", data=adata.X.indptr.astype(np.int32))
        grp.create_dataset("shape", data=np.array([n_genes, n_cells], dtype=np.int64))

        # features
        ftrs = grp.create_group("features")
        ftrs.create_dataset(
            "id", data=np.array(gene_ids, dtype=f"|S{max(map(len, gene_ids))}")
        )
        ftrs.create_dataset(
            "name",
            data=np.array(adata.var_names, dtype=f"|S{max(map(len, adata.var_names))}"),
        )
        ftrs.create_dataset(
            "feature_type",
            data=np.array(feature_types, dtype=f"|S{max(map(len, feature_types))}"),
        )
        ftrs.create_dataset(
            "genome", data=np.array(genomes, dtype=f"|S{max(map(len, genomes))}")
        )

        # Visium-specific metadata
        all_tag_keys = np.array(
            [b"feature_type", b"genome", b"id", b"name", b"library_ids"]
        )
        ftrs.create_dataset("_all_tag_keys", data=all_tag_keys)

        library_ids_dataset = np.array([library_id.encode()] * n_genes)
        ftrs.create_dataset("library_ids", data=library_ids_dataset)

        # HDF5 attributes required by scanpy.read_visium
        f.attrs["library_ids"] = np.array([library_id.encode()])
        f.attrs["chemistry_description"] = "Spatial3"
        f.attrs["software_version"] = "CustomPythonScript"



import argparse
import pysam


def main():
    parser = argparse.ArgumentParser("Add barcode and UMI tag to celescope BAM")
    parser.add_argument("--bam", help="input bam", required=True)
    parser.add_argument("--outdir", help="output directory", default="./")
    args = parser.parse_args()
    add_tag(args.bam, args.outdir)


def add_tag(bam, outdir):
    bam_out = f"{outdir}/add_tag.bam"
    handle_in = pysam.AlignmentFile(bam, "rb")
    header = handle_in.header
    handle_out = pysam.AlignmentFile(bam_out, "wb", header=header)
    for read in handle_in:
        attr = read.query_name.split(":")
        barcode = attr[0]
        umi = attr[1]
        read.set_tag(tag="CB", value=barcode, value_type="Z")
        read.set_tag(tag="UB", value=umi, value_type="Z")
        handle_out.write(read)
    handle_in.close()
    handle_out.close()


if __name__ == "__main__":
    main()



from celescope.space.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_space(Multi):
    def starsolo(self, sample):
        step = "starsolo"
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq1 {arr["fq1_str"]} --fq2 {arr["fq2_str"]}  --spatial {self.col4_dict[sample]} '
        )
        self.process_cmd(
            cmd,
            step,
            sample,
            m=int(self.args.limitBAMsortRAM / 1e9),
            x=self.args.thread,
        )

    def analysis(self, sample):
        step = "analysis"
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f"{cmd_line} "
            f"--raw {self.outdir_dic[sample]['outs']}/raw "
            f"--filtered {self.outdir_dic[sample]['outs']}/filtered "
            f"--spatial {self.col4_dict[sample]} "
        )
        self.process_cmd(cmd, step, sample)


def main():
    multi = Multi_space(__ASSAY__, min_col=4, pair=(1, 2))
    multi.run()


if __name__ == "__main__":
    main()



from celescope.probe.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_probe(Multi):
    """
    probe and primer detection
    """


def main():
    multi = Multi_probe(__ASSAY__)
    multi.run()


if __name__ == "__main__":
    main()



STEPS = [
    "mkref",
    "sample",
    "starsolo",
    "target_metrics",
    "variant_calling",
    "filter_snp",
    "analysis_snp",
]
__ASSAY__ = "snp"

PANEL = {"lung_1", "blood_1", "CHIP"}

REMOVE_FROM_MULTI = {
    "mkref",
}



import unittest
from collections import Counter

import pysam

import celescope.tools.parse_chemistry as parse_chemistry
from celescope.__init__ import HELP_DICT
from celescope.chemistry_dict import chemistry_dict
from celescope.tools import utils
from celescope.tools.step import Step, s_common

MAX_TRUST4_READ_PER_BARCODE = 80000


class Barcode(Step):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title)
        self.fq1_list = args.fq1.split(",")
        self.fq2_list = args.fq2.split(",")
        self.chemistry = parse_chemistry.get_chemistry(
            self.assay, args.chemistry, self.fq1_list
        )
        self.pattern_dict, self.bc = parse_chemistry.get_pattern_dict_and_bc(
            self.chemistry, args.pattern, args.whitelist
        )

        self.raw_list, self.mismatch_list = (
            parse_chemistry.create_mismatch_origin_dicts_from_whitelists(self.bc, 1)
        )
        # v3
        self.offset_runner = parse_chemistry.AutoRNA(self.fq1_list)
        # output
        self.out_fq2 = f"{self.outdir}/{self.sample}_2.fq"
        self.isflv = self.assay == "flv_trust4"
        if self.isflv:
            self.out_fq1 = f"{self.outdir}/{self.sample}_1.fq"
            self.match_barcodes = set(
                utils.get_barcode_from_match_dir(args.match_dir)[0]
            )

    def get_bc_umi(self, seq):
        if self.chemistry == "GEXSCOPE-V3":
            offset = self.offset_runner.v3_offset(seq)
            seq = seq[offset:]
        elif self.chemistry == "flv_rna-V2":
            offset = self.offset_runner.flv_rna_v2_offset(seq)
            seq = seq[offset:]
        bc_list = [seq[x] for x in self.pattern_dict["C"]]
        if self.isflv:
            bc_list = [utils.reverse_complement(bc) for bc in bc_list[::-1]]
        valid, corrected, corrected_seq = parse_chemistry.check_seq_mismatch(
            bc_list, self.raw_list, self.mismatch_list
        )
        if not valid:
            umi = None
        else:
            umi = seq[self.pattern_dict["U"][0]]
        return valid, corrected, corrected_seq, umi

    @utils.add_log
    def run(self):
        raw_reads = valid_reads = corrected_reads = 0
        # quality
        cb_quality_counter = Counter()
        umi_quality_counter = Counter()
        if self.isflv:
            out_fh1 = utils.generic_open(self.out_fq1, "wt")
            bc_read_counter = Counter()
            match_reads = 0

        with utils.generic_open(self.out_fq2, "wt") as out_fh2:
            for fq1, fq2 in zip(self.fq1_list, self.fq2_list):
                fq1 = pysam.FastxFile(fq1)
                fq2 = pysam.FastxFile(fq2)
                for e1, e2 in zip(fq1, fq2):
                    raw_reads += 1
                    valid, corrected, corrected_seq, umi = self.get_bc_umi(e1.sequence)
                    if valid:
                        valid_reads += 1
                        if corrected:
                            corrected_reads += 1
                        read_name = f"{corrected_seq}:{umi}:{raw_reads}"
                        if not self.isflv:
                            out_fh2.write(
                                utils.fastq_line(read_name, e2.sequence, e2.quality)
                            )  # type: ignore
                        else:
                            if corrected_seq not in self.match_barcodes:
                                continue
                            match_reads += 1
                            bc_read_counter[corrected_seq] += 1
                            seq1 = corrected_seq + umi
                            qual1 = "F" * len(seq1)
                            if (
                                bc_read_counter[corrected_seq]
                                <= MAX_TRUST4_READ_PER_BARCODE
                            ):
                                out_fh1.write(utils.fastq_line(read_name, seq1, qual1))
                                out_fh2.write(
                                    utils.fastq_line(read_name, e2.sequence, e2.quality)
                                )  # type: ignore

                    cb_quality_counter.update(
                        "".join([e1.quality[slice] for slice in self.pattern_dict["C"]])  # type: ignore
                    )
                    umi_quality_counter.update(
                        "".join([e1.quality[self.pattern_dict["U"][0]]])  # type: ignore
                    )

        self.add_metric(
            name="Raw Reads",
            value=raw_reads,
            help_info="total reads from FASTQ files",
        )
        self.add_metric(
            name="Valid Reads",
            value=valid_reads,
            total=raw_reads,
            help_info="reads with correct barcode",
        )
        self.add_metric(
            name="Corrected Reads",
            value=corrected_reads,
            total=raw_reads,
            help_info="Reads with barcodes that are not in the whitelist but are within one Hamming distance of it",
        )
        if not cb_quality_counter:
            q30_cb = q30_umi = 0
        else:
            q30_cb = sum(
                [
                    cb_quality_counter[k]
                    for k in cb_quality_counter
                    if Barcode.chr_to_int(k) >= 30
                ]
            ) / float(sum(cb_quality_counter.values()))
            q30_umi = sum(
                [
                    umi_quality_counter[k]
                    for k in umi_quality_counter
                    if Barcode.chr_to_int(k) >= 30
                ]
            ) / float(sum(umi_quality_counter.values()))
        self.add_metric(
            name="Q30 of Barcode",
            value=f"{round(q30_cb * 100,2)}%",
            help_info="percent of barcode base pairs with quality scores over Q30",
        )
        self.add_metric(
            name="Q30 of UMI",
            value=f"{round(q30_umi * 100, 2)}%",
            help_info="percent of UMI base pairs with quality scores over Q30",
        )

        if self.assay == "flv_trust4":
            self.add_metric(
                name="Valid Matched Reads",
                value=match_reads,
                total=valid_reads,
                help_info="reads match with flv_rna cell barcodes",
            )

            self.add_metric(
                name="Matched Barcodes",
                value=len(bc_read_counter),
                help_info="barcodes match with flv_rna",
            )

    @staticmethod
    def chr_to_int(chr, offset=33):
        """Convert Phred quality to int"""
        return ord(chr) - offset


@utils.add_log
def barcode(args):
    with Barcode(args, display_title="Demultiplexing") as runner:
        runner.run()


def get_opts_barcode(parser, sub_program=True):
    parser.add_argument(
        "--chemistry",
        help=HELP_DICT["chemistry"],
        choices=list(chemistry_dict.keys()),
        default="auto",
    )
    parser.add_argument(
        "--pattern",
        help="""The pattern of R1 reads, e.g. `C8L16C8L16C8L1U12T18`. The number after the letter represents the number 
        of bases.  
- `C`: cell barcode  
- `L`: linker(common sequences)  
- `U`: UMI    
- `T`: poly T""",
    )
    parser.add_argument(
        "--whitelist",
        help="Cell barcode whitelist file path, one cell barcode per line.",
    )
    if sub_program:
        parser.add_argument(
            "--fq1",
            help="R1 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--fq2",
            help="R2 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--match_dir", help="Matched scRNA-seq directory, required for flv_trust4"
        )
        parser = s_common(parser)

    return parser


if __name__ == "__main__":
    unittest.main()



from celescope.dynaseq.__init__ import __ASSAY__
from celescope.rna.multi_rna import Multi_rna
from celescope.tools.__init__ import (
    FILTERED_MATRIX_DIR_SUFFIX,
    BARCODE_FILE_NAME,
    FEATURE_FILE_NAME,
    STARSOLO_BAM_SUFFIX,
    OUTS_DIR,
)


class Multi_dynaseq(Multi_rna):
    """
    ## Usage

    ```
        multi_dynaseq\\
        --mapfile ./rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus\\
    ```

    For control sample, set --control to skip replacement step.
    ```
        multi_dynaseq\\
        --mapfile ./rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus\\
        --control
    ```
    """

    def starsolo(self, sample):
        step = "starsolo"
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq1 {arr["fq1_str"]} --fq2 {arr["fq2_str"]} '
            f'--STAR_param "--outFilterScoreMinOverLread 0.3 --outFilterMatchNminOverLread 0.3" '
            f'--SAM_attributes MD '
        )
        self.process_cmd(
            cmd,
            step,
            sample,
            m=int(self.args.limitBAMsortRAM / 1e9),
            x=self.args.thread,
        )

    def conversion(self, sample):
        step = "conversion"
        bam = f'{self.outdir_dic[sample]["outs"]}/{sample}_{STARSOLO_BAM_SUFFIX}'
        star_log = f'{self.outdir_dic[sample]["starsolo"]}/{sample}_Log.final.out'
        cell = f'{self.outdir_dic[sample]["outs"]}/{FILTERED_MATRIX_DIR_SUFFIX}/{BARCODE_FILE_NAME}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = f"{cmd_line} " f"--bam {bam} " f"--cell {cell} " f"--star_log {star_log} "
        self.process_cmd(
            cmd, step, sample, m=self.args.conversionMem, x=self.args.thread
        )

    def substitution(self, sample):
        step = "substitution"
        bam = f'{self.outdir_dic[sample]["conversion"]}/{sample}.PosTag.bam'
        snp = f'{self.outdir_dic[sample]["conversion"]}/{sample}.snp.csv'
        cmd_line = self.get_cmd_line(step, sample)
        bg_para = ""
        if sample in self.col5_dict:
            bg_para = f"--bg {self.col5_dict[sample]} "
        cmd = f"{cmd_line} " f"--bam {bam} " f"--bg {snp} {bg_para} "
        self.process_cmd(cmd, step, sample, m=1, x=1)

    def replacement(self, sample):
        step = "replacement"
        bam = f'{self.outdir_dic[sample]["conversion"]}/{sample}.PosTag.bam'
        snp = f'{self.outdir_dic[sample]["conversion"]}/{sample}.snp.csv'
        tsne_file = f"{self.outdir_dic[sample][OUTS_DIR]}/tsne_coord.tsv"
        cell = f'{self.outdir_dic[sample]["outs"]}/{FILTERED_MATRIX_DIR_SUFFIX}/{BARCODE_FILE_NAME}'
        gene = f'{self.outdir_dic[sample]["outs"]}/{FILTERED_MATRIX_DIR_SUFFIX}/{FEATURE_FILE_NAME}'
        cmd_line = self.get_cmd_line(step, sample)
        bg_para = ""
        if sample in self.col5_dict:
            bg_para = f"--bg {self.col5_dict[sample]} "
        cmd = (
            f"{cmd_line} "
            f"--bam {bam} "
            f"--bg {snp} {bg_para} "
            f"--tsne {tsne_file} "
            f"--cell {cell} "
            f"--gene {gene} "
        )
        self.process_cmd(
            cmd, step, sample, m=self.args.replacementMem, x=self.args.thread
        )


def main():
    multi = Multi_dynaseq(__ASSAY__)
    multi.run()


if __name__ == "__main__":
    main()



from celescope.tools import utils
from celescope.tools import reference


class Mkgtf:
    """
    ##Features

    - Filter GTF files.

    GTF files downloaded from sites like ENSEMBL and UCSC often contain transcripts and genes which need to be filtered.
    Celescope provides mkgtf, a simple utility to filter genes. The command syntax requires input and output fitlered GTF file names.

    The following gene biotypes will be kept.
    ```
    protein_coding
    lncRNA
    antisense
    IG_LV_gene
    IG_V_gene
    IG_V_pseudogene
    IG_D_gene
    IG_J_gene
    IG_J_pseudogene
    IG_C_gene
    IG_C_pseudogene
    TR_V_gene
    TR_V_pseudogene
    TR_D_gene
    TR_J_gene
    TR_J_pseudogene
    TR_C_gene
    ```
    """

    def __init__(self, in_gtf, out_gtf, attributes, add_intron=True):
        self.in_gtf_fn = in_gtf
        self.out_gtf_fn = out_gtf
        self.add_intron = add_intron

        self.attributes = {}
        for attr_str in attributes.split(";"):
            if attr_str:
                attr, val = attr_str.split("=")
                val = set(val.split(","))
                self.attributes[attr] = val

    @utils.add_log
    def run(self):
        runner = reference.GtfBuilder(
            self.in_gtf_fn, self.out_gtf_fn, self.attributes, add_intron=self.add_intron
        )
        runner.build_gtf()


def mkgtf(args):
    runner = Mkgtf(args.gtf, args.out_gtf, args.attributes, not args.skip_intron)
    runner.run()


def get_opts_mkgtf(parser, sub_program=True):
    if sub_program:
        parser.add_argument("gtf", help="raw gtf file")
        parser.add_argument("out_gtf", help="output gtf file")
        parser.add_argument(
            "--attributes",
            help="Attributes to keep. Example: `gene_biotype=protein_coding,lncRNA,antisense;`",
            default="gene_biotype=protein_coding,lncRNA,antisense,IG_LV_gene,IG_V_gene,IG_V_pseudogene,IG_D_gene,IG_J_gene,IG_J_pseudogene,IG_C_gene,IG_C_pseudogene,TR_V_gene,TR_V_pseudogene,TR_D_gene,TR_J_gene,TR_J_pseudogene,TR_C_gene;",
        )
        parser.add_argument(
            "--skip_intron", action="store_true", help="Do not add intron to gtf"
        )

    return parser



import argparse
import json
import os
from collections import defaultdict

import pandas as pd
from celescope.tools import utils


@utils.add_log
def append_sample_data(sample, sample_data_dict, all_data_dict, steps):
    """
    append sample data to all_data_dict
    Args:
        all_data_dict - key: step, value: list of sample series
    """
    for step in steps:
        title = step + "_summary"
        if title in sample_data_dict:
            metric_list = (
                sample_data_dict[title]["metric_list"]
                + sample_data_dict[title]["comment_metric_list"]
            )
            if metric_list:
                metric_dict = {"sample": sample}
                for metric in metric_list:
                    name = metric["name"].replace(" ", "_")
                    metric_dict[name] = metric["display"]

                all_data_dict[step].append(pd.Series(metric_dict))


@utils.add_log
def write_merge_report(all_data_dict, merge_report_handle, steps):
    """
    merge_report:
        ## sample_summary
        sample,sample_ID,Assay...
        sample_1,sample_1,rna...
        ## barcode_summary

    Args:
        all_data_dict - key: title, value: list of sample series
    """

    for step in steps:
        if all_data_dict[step]:
            merge_report_handle.write(f"## {step}_summary\n")
            df = pd.concat(all_data_dict[step], axis=1).T
            df.to_csv(merge_report_handle, index=False, mode="a", sep="\t")
            merge_report_handle.write("\n")


@utils.add_log
def run(args):
    # out
    os.chdir(args.outdir)
    out_file = "merge.xls"

    all_data_dict = defaultdict(list)
    steps = args.steps.split(",")
    samples = args.samples.split(",")
    run.logger.info(f"samples: {samples}")

    for sample in samples:
        sample_data_dict = json.load(open(f"{sample}/.data.json", "r"))
        append_sample_data(sample, sample_data_dict, all_data_dict, steps)

    write_merge_report(all_data_dict, open(out_file, "w"), steps)


def main():
    parser = argparse.ArgumentParser("merge report")
    parser.add_argument("--outdir", help="outdir", required=True)
    parser.add_argument("--samples", help="samples, seperated by comma", required=True)
    parser.add_argument("--steps", help="steps", required=True)
    parser.add_argument(
        "--rm_files", action="store_true", help="remove all fq and bam after running"
    )
    args = parser.parse_args()

    run(args)


@utils.add_log
def rm_files():
    cmd = """
        find . -iname '*.fq*' -delete;
        find . -iname '*.bam' -not -path './*/*.featureCounts/*name_sorted.bam' -delete;
    """
    os.system(cmd)


if __name__ == "__main__":
    main()



"""Split fastq file according to barcode. Append UMI to read name with umi_separator semicolon ':'"""

import io
from celescope.tools.step import Step, s_common
from celescope.tools import utils
import celescope.tools.parse_chemistry as parse_chemistry
import pysam
import sys
from celescope.chemistry_dict import chemistry_dict
from celescope.__init__ import HELP_DICT
from celescope.bulk_rna.starsolo import get_barcode_sample


UMI_SEPARATOR = ":"


class Split_fastq(Step):
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.fq1_list = args.fq1.split(",")
        self.fq2_list = args.fq2.split(",")

        chemistry = parse_chemistry.get_chemistry(
            self.assay, args.chemistry, self.fq1_list
        )
        self.pattern_dict, bc = parse_chemistry.get_pattern_dict_and_bc(
            chemistry, args.pattern, args.whitelist
        )
        self.raw_list, self.mismatch_list = (
            parse_chemistry.create_mismatch_origin_dicts_from_whitelists(bc, 1)
        )

        self.barcode_sample = get_barcode_sample(bc[0], args.well_sample)

    @utils.add_log
    def split(self):
        self.fh_dict = {}
        for barcode, sample in self.barcode_sample.items():
            file_obj = utils.generic_open(
                f"{self.outdir}/{sample}.fq.gz", "wb", compresslevel=1
            )
            self.fh_dict[barcode] = io.BufferedWriter(
                file_obj, buffer_size=16 * 1024 * 1024
            )

        for fq1, fq2 in zip(self.fq1_list, self.fq2_list):
            fq1 = pysam.FastxFile(fq1)
            fq2 = pysam.FastxFile(fq2)
            for e1, e2 in zip(fq1, fq2):
                bc_list = [e1.sequence[x] for x in self.pattern_dict["C"]]
                valid, corrected, corrected_bc = parse_chemistry.check_seq_mismatch(
                    bc_list, self.raw_list, self.mismatch_list
                )
                if valid and corrected_bc in self.barcode_sample:
                    umi = e1.sequence[self.pattern_dict["U"][0]]
                    name = e1.name + UMI_SEPARATOR + umi
                    self.fh_dict[corrected_bc].write(
                        utils.fastq_line(name, e2.sequence, e2.quality).encode()
                    )

    def run(self):
        self.split()


@utils.add_log
def split_fastq(args):
    if not args.split_fastq:
        sys.stderr.write("--split_fastq not set. Skip split_fastq\n")
        return
    split_fastq_obj = Split_fastq(args)
    split_fastq_obj.run()


def get_opts_split_fastq(parser, sub_program):
    parser.add_argument(
        "--chemistry",
        help=HELP_DICT["chemistry"],
        choices=list(chemistry_dict.keys()),
        default="auto",
    )
    parser.add_argument(
        "--pattern",
        help="""The pattern of R1 reads, e.g. `C8L16C8L16C8L1U12T18`. The number after the letter represents the number 
        of bases.  
        - `C`: cell barcode  
        - `L`: linker(common sequences)  
        - `U`: UMI    
        - `T`: poly T""",
    )
    parser.add_argument(
        "--whitelist",
        help="Cell barcode whitelist file path, one cell barcode per line.",
    )
    parser.add_argument(
        "--well_sample",
        help="tsv file of well numbers and sample names. The first column is well numbers and the second column is sample names.",
        required=True,
    )
    parser.add_argument(
        "--split_fastq",
        help="Split fastq file according to well barcodes. Append UMI to read name with umi_separator semicolon ':'",
        action="store_true",
    )
    if sub_program:
        parser.add_argument(
            "--fq1",
            help="R1 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--fq2",
            help="R2 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--barcodes",
            help="barcode file.",
            required=True,
        )
        s_common(parser)
    return parser



import unittest
from collections import Counter, defaultdict
import sys

import pysam
import pandas as pd

import celescope.tools.parse_chemistry as parse_chemistry
from celescope.__init__ import HELP_DICT
from celescope.chemistry_dict import chemistry_dict
from celescope.tools import utils
from celescope.tools.step import Step, s_common


SEQUENCE_LENGTH = 150
PRIMER_OFFSET = 5


@utils.add_log
def fasta_to_dict(fasta, max_mismatch):
    seq_dict, _ = utils.read_fasta(fasta)
    mismatch_dict = {}
    length_dict = {}
    for seq_id, seq in seq_dict.items():
        mismatch_dict[seq_id] = parse_chemistry.create_mismatch_seqs(
            seq, max_mismatch=max_mismatch
        )
        length_dict[seq_id] = len(seq)

    return mismatch_dict, length_dict


@utils.add_log
def get_metrics_df(read_counter, umi_counter, total_reads, total_umis):
    all_probes = read_counter.keys()
    df = pd.DataFrame(index=sorted(all_probes))
    df["read"] = df.index.map(lambda x: read_counter.get(x, 0))
    df["umi"] = df.index.map(lambda x: umi_counter.get(x, 0))
    df["read_percent"] = df["read"] / total_reads * 100
    df["umi_percent"] = df["umi"] / total_umis * 100

    df["read_percent"] = df["read_percent"].round(2)
    df["umi_percent"] = df["umi_percent"].round(2)
    df = df.reset_index().rename(columns={"index": "id"})
    df = df.sort_values(by="read", ascending=False)
    return df


class Barcode(Step):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title)
        self.fq1_list = args.fq1.split(",")
        self.fq2_list = args.fq2.split(",")
        self.chemistry = parse_chemistry.get_chemistry(
            self.assay, args.chemistry, self.fq1_list
        )
        self.pattern_dict, self.bc = parse_chemistry.get_pattern_dict_and_bc(
            self.chemistry, args.pattern, args.whitelist
        )
        self.bc_umi_length = sum(
            s.stop - s.start for s_list in self.pattern_dict.values() for s in s_list
        )
        sys.stderr.write(f"bc+umi+linker length: {self.bc_umi_length}\n")

        self.raw_list, self.mismatch_list = (
            parse_chemistry.create_mismatch_origin_dicts_from_whitelists(self.bc, 1)
        )
        # v3
        self.offset_runner = parse_chemistry.AutoRNA(self.fq1_list)
        #
        self.probe_mismatch_dict = {}
        if args.probe_fasta:
            self.probe_mismatch_dict, self.probe_length_dict = fasta_to_dict(
                args.probe_fasta, args.max_mismatch
            )

        self.primer_mismatch_dict = {}
        if args.primer_fasta:
            self.primer_mismatch_dict, self.primer_length_dict = fasta_to_dict(
                args.primer_fasta, args.max_mismatch
            )

        # output
        self.probe_count_file = f"{self.out_prefix}_probe_count.tsv"
        self.primer_count_file = f"{self.out_prefix}_primer_count.tsv"
        self.outs = [self.probe_count_file, self.primer_count_file]

    def get_bc_umi(self, seq):
        if self.chemistry == "GEXSCOPE-V3":
            offset = self.offset_runner.v3_offset(seq)
            seq = seq[offset:]
        elif self.chemistry == "flv_rna-V2":
            offset = self.offset_runner.flv_rna_v2_offset(seq)
            seq = seq[offset:]
        bc_list = [seq[x] for x in self.pattern_dict["C"]]
        if self.chemistry == "flv":
            bc_list = [utils.reverse_complement(bc) for bc in bc_list[::-1]]
        valid, corrected, corrected_seq = parse_chemistry.check_seq_mismatch(
            bc_list, self.raw_list, self.mismatch_list
        )
        if not valid:
            umi = None
        else:
            umi = seq[self.pattern_dict["U"][0]]
        return valid, corrected, corrected_seq, umi

    @utils.add_log
    def run(self):
        raw_reads = valid_reads = corrected_reads = 0
        # quality
        cb_quality_counter = Counter()
        umi_quality_counter = Counter()
        all_umi_set = set()

        probe_read_counter = Counter()
        primer_read_counter = Counter()
        probe_umi_set = defaultdict(set)
        primer_umi_set = defaultdict(set)

        for fq1, fq2 in zip(self.fq1_list, self.fq2_list):
            fq1 = pysam.FastxFile(fq1)
            fq2 = pysam.FastxFile(fq2)
            for e1, e2 in zip(fq1, fq2):
                raw_reads += 1
                valid, corrected, corrected_seq, umi = self.get_bc_umi(e1.sequence)
                if valid:
                    valid_reads += 1
                    if corrected:
                        corrected_reads += 1
                    all_umi_set.add((corrected_seq, umi))

                    for probe_id, seqs in self.probe_mismatch_dict.items():
                        seq_length = self.probe_length_dict[probe_id]
                        for start in range(
                            self.bc_umi_length, SEQUENCE_LENGTH - seq_length
                        ):
                            if e1.sequence[start : start + seq_length] in seqs:
                                probe_read_counter[probe_id] += 1
                                probe_umi_set[probe_id].add((corrected_seq, umi))
                                break

                    for primer_id, seqs in self.primer_mismatch_dict.items():
                        seq_length = self.primer_length_dict[primer_id]
                        for start in range(0, PRIMER_OFFSET):
                            if e2.sequence[start : start + seq_length] in seqs:
                                primer_read_counter[primer_id] += 1
                                primer_umi_set[primer_id].add((corrected_seq, umi))
                                break

                cb_quality_counter.update(
                    "".join([e1.quality[slice] for slice in self.pattern_dict["C"]])  # type: ignore
                )
                umi_quality_counter.update(
                    "".join([e1.quality[self.pattern_dict["U"][0]]])  # type: ignore
                )

        total_umi = len(all_umi_set)

        probe_umi_counter = {k: len(v) for k, v in probe_umi_set.items()}
        probe_df = get_metrics_df(
            probe_read_counter, probe_umi_counter, valid_reads, total_umi
        )
        probe_df.to_csv(self.probe_count_file, sep="\t", index=False)
        self.add_table(
            title="Probe",
            table_id="table_probe",
            df=probe_df,
        )

        primer_umi_counter = {k: len(v) for k, v in primer_umi_set.items()}
        primer_df = get_metrics_df(
            primer_read_counter, primer_umi_counter, valid_reads, total_umi
        )
        primer_df.to_csv(self.primer_count_file, sep="\t", index=False)
        self.add_table(
            title="Primer",
            table_id="table_primer",
            df=primer_df,
        )

        self.add_metric(
            name="Raw Reads",
            value=raw_reads,
            help_info="total reads from FASTQ files",
        )
        self.add_metric(
            name="Valid Reads",
            value=valid_reads,
            total=raw_reads,
            help_info="reads with correct barcode",
        )
        self.add_metric(
            name="Corrected Reads",
            value=corrected_reads,
            total=raw_reads,
            help_info="Reads with barcodes that are not in the whitelist but are within one Hamming distance of it",
        )
        self.add_metric(
            name="Total UMI",
            value=total_umi,
            help_info="total UMI count",
        )
        q30_cb = sum(
            [
                cb_quality_counter[k]
                for k in cb_quality_counter
                if Barcode.chr_to_int(k) >= 30
            ]
        ) / float(sum(cb_quality_counter.values()))
        q30_umi = sum(
            [
                umi_quality_counter[k]
                for k in umi_quality_counter
                if Barcode.chr_to_int(k) >= 30
            ]
        ) / float(sum(umi_quality_counter.values()))
        self.add_metric(
            name="Q30 of Barcode",
            value=f"{round(q30_cb * 100,2)}%",
            help_info="percent of barcode base pairs with quality scores over Q30",
        )
        self.add_metric(
            name="Q30 of UMI",
            value=f"{round(q30_umi * 100, 2)}%",
            help_info="percent of UMI base pairs with quality scores over Q30",
        )

    @staticmethod
    def chr_to_int(chr, offset=33):
        """Convert Phred quality to int"""
        return ord(chr) - offset


@utils.add_log
def barcode(args):
    with Barcode(args, display_title="Demultiplexing") as runner:
        runner.run()


def get_opts_barcode(parser, sub_program=True):
    parser.add_argument(
        "--chemistry",
        help=HELP_DICT["chemistry"],
        choices=list(chemistry_dict.keys()),
        default="auto",
    )
    parser.add_argument(
        "--pattern",
        help="""The pattern of R1 reads, e.g. `C8L16C8L16C8L1U12T18`. The number after the letter represents the number 
        of bases.  
- `C`: cell barcode  
- `L`: linker(common sequences)  
- `U`: UMI    
- `T`: poly T""",
    )
    parser.add_argument(
        "--whitelist",
        help="Cell barcode whitelist file path, one cell barcode per line.",
    )
    parser.add_argument("--probe_fasta", help="Probe fasta file.")
    parser.add_argument("--primer_fasta", help="Primer fasta file.")
    parser.add_argument(
        "--max_mismatch",
        help="Maximum mismatch in probe in primer sequence.",
        default=2,
        type=int,
    )
    if sub_program:
        parser.add_argument(
            "--fq1",
            help="R1 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--fq2",
            help="R2 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--match_dir", help="Matched scRNA-seq directory, required for flv_trust4"
        )
        parser = s_common(parser)

    return parser


if __name__ == "__main__":
    unittest.main()



"""
vdj mapping
"""

import pandas as pd
import subprocess
import pysam
from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.vdj.__init__ import CHAINS
from xopen import xopen


SPLIT_N_CHUNKS = 4


class Mapping_vdj(Step):
    """
    ## Features
    - Align R2 reads to IGMT(http://www.imgt.org/) database sequences with blast.

    ## Output
    - `{sample}_airr.tsv` The alignment result of each UMI.
    A tab-delimited file compliant with the AIRR Rearrangement schema(https://docs.airr-community.org/en/stable/datarep/rearrangements.html)

    - `{sample}_UMI_count_unfiltered.tsv` UMI reading for each (barcode, chain, VJ_pair) combination.

    - `{sample}_UMI_count_filtered.tsv` For each (barcode, chain) combination, only the record with the
    most VJ_pair UMI reads is kept.

    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.ref_path = args.ref_path
        self.seqtype = args.type
        self.species = args.species
        self.split_fasta = args.split_fasta
        self.chains = CHAINS[self.seqtype]

        # out
        self.UMI_count_unfiltered_file = f"{self.out_prefix}_UMI_count_unfiltered.tsv"
        self.UMI_count_filtered_file = f"{self.out_prefix}_UMI_count_filtered.tsv"

        if self.split_fasta:
            utils.check_mkdir(f"{self.outdir}/tmp/")
            self.UMI_number = int(
                self.get_slot_key(
                    slot="metrics", step_name="consensus", key="UMI Counts"
                )
            )
            self.tmp_fasta = [
                f"{self.outdir}/tmp/{self.sample}_{i}.fasta"
                for i in range(SPLIT_N_CHUNKS)
            ]
            self.airr_out = [
                f"{self.out_prefix}_airr_{i}.tsv" for i in range(SPLIT_N_CHUNKS)
            ]

        else:
            self.airr_out = [f"{self.out_prefix}_airr.tsv"]

    def run(self):
        # run igblstn
        if not self.split_fasta:
            self.igblast(fasta=self.args.fasta, airr_out=self.airr_out[0])
        else:
            self.split_fasta_file()
            for i in range(len(self.tmp_fasta)):
                self.igblast(fasta=self.tmp_fasta[i], airr_out=self.airr_out[i])

        self.mapping_summary()

    @utils.add_log
    def split_fasta_file(self):
        fh_tmp_fasta = [xopen(i, "w") for i in self.tmp_fasta]
        umi_count = 0
        chunk = self.UMI_number // SPLIT_N_CHUNKS
        with pysam.FastxFile(self.args.fasta) as f:
            for read in f:
                umi_count += 1
                if umi_count <= chunk:
                    fh_tmp_fasta[0].write(utils.fasta_line(read.name, read.sequence))
                elif chunk < umi_count <= 2 * chunk:
                    fh_tmp_fasta[1].write(utils.fasta_line(read.name, read.sequence))
                elif 2 * chunk < umi_count <= 3 * chunk:
                    fh_tmp_fasta[2].write(utils.fasta_line(read.name, read.sequence))
                else:
                    fh_tmp_fasta[3].write(utils.fasta_line(read.name, read.sequence))

        for i in fh_tmp_fasta:
            i.close()

    @utils.add_log
    def igblast(self, fasta, airr_out):
        if self.seqtype == "TCR":
            chain, ig_seqtype = "TR", "TCR"
        else:
            chain, ig_seqtype = "IG", "Ig"

        cmd = (
            f"igblastn -query {fasta} "
            f"-organism {self.species} "
            f"-ig_seqtype {ig_seqtype} "
            f"-auxiliary_data optional_file/{self.species}_gl.aux "
            f"-num_threads {self.args.thread} "
            f"-germline_db_V {self.ref_path}/{chain}V.fa "
            f"-germline_db_D {self.ref_path}/{chain}D.fa "
            f"-germline_db_J {self.ref_path}/{chain}J.fa "
            "-domain_system imgt -show_translation -outfmt 19 "  # outfmt19 is an AIRR tab-delimited file, IgBLAST v1.9.0 or higher required.
            f"-out {airr_out} "
        )

        self.igblast.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def mapping_summary(self):
        self.add_metric(name="Species", value=self.species, help_info="Human or Mouse")

        df_total_confident = pd.DataFrame()
        (
            total_reads,
            map_to_any_vdj_gene_num,
            cdr3_num,
            correct_cdr3_num,
            confident_num,
        ) = 0, 0, 0, 0, 0

        for df in self.airr_out:
            df = pd.read_csv(df, sep="\t")
            df.fillna("", inplace=True)
            total_reads += df.shape[0]

            # mapping to any vdj genes
            df = df[(df["v_call"] != "") | (df["d_call"] != "") | (df["j_call"] != "")]
            map_to_any_vdj_gene_num += df.shape[0]
            # UMIs with CDR3
            df = df[df["cdr3_aa"] != ""]
            cdr3_num += df.shape[0]
            # UMIs with Correct CDR3
            df = df[
                ~(df["cdr3_aa"].str.contains(r"\*"))
                & ~(df["cdr3_aa"].str.contains("X"))
            ]
            correct_cdr3_num += df.shape[0]
            # UMIs Mapped Confidently To VJ Gene
            df = df[df["productive"] == "T"]
            confident_num += df.shape[0]

            df_total_confident = pd.concat([df_total_confident, df])

        self.add_metric(
            name="UMIs Mapped to Any VDJ Gene",
            value=map_to_any_vdj_gene_num,
            total=total_reads,
            help_info="UMI Mapped to any germline VDJ gene segments",
        )
        self.add_metric(
            name="UMIs with CDR3",
            value=cdr3_num,
            total=total_reads,
            help_info="UMIs with CDR3 sequence",
        )
        self.add_metric(
            name="UMIs with Correct CDR3",
            value=correct_cdr3_num,
            total=total_reads,
            help_info="UMIs with CDR3 might have stop codon and these UMIs(or Reads) are classified as incorrect",
        )
        self.add_metric(
            name="UMIs Mapped Confidently to VJ Gene",
            value=confident_num,
            total=total_reads,
            help_info="UMIs with productive rearrangement mapped to VJ gene pairs and with correct CDR3",
        )

        # UMIs Mapped Confidently to each chain
        for chain in self.chains:
            df_chain = df_total_confident[df_total_confident.locus == chain]
            self.add_metric(
                name=f"UMIs Mapped to {chain}",
                value=df_chain.shape[0],
                total=total_reads,
                help_info=f"UMIs mapped confidently to {chain}",
            )

        # output file
        df_VJ = df_total_confident[
            [
                "sequence_id",
                "locus",
                "v_call",
                "d_call",
                "j_call",
                "junction",
                "junction_aa",
            ]
        ]
        df_VJ.rename(
            columns={
                "sequence_id": "readID",
                "locus": "chain",
                "v_call": "bestVGene",
                "d_call": "bestDGene",
                "j_call": "bestJGene",
                "junction": "nSeqCDR3",
                "junction_aa": "aaSeqCDR3",
            },
            inplace=True,
        )
        df_VJ["barcode"] = df_VJ["readID"].apply(lambda x: x.split(":")[0])
        df_VJ["UMI"] = df_VJ["readID"].apply(lambda x: x.split(":")[1])

        for i in ["bestVGene", "bestDGene", "bestJGene"]:
            df_VJ[i] = df_VJ[i].apply(lambda x: x.split("*")[0])

        # filter1: keep top 1 in each combinations
        groupby_elements = [
            "barcode",
            "chain",
            "bestVGene",
            "bestJGene",
            "aaSeqCDR3",
            "nSeqCDR3",
        ]

        # unique UMI
        df_UMI = df_VJ.drop_duplicates(subset=["barcode", "UMI"], keep="first")
        df_UMI_count = df_UMI.groupby(groupby_elements, as_index=False).agg(
            {"UMI": "count"}
        )
        df_UMI_count = df_UMI_count.sort_values("UMI", ascending=False)
        # out unfiltered
        df_UMI_count.to_csv(self.UMI_count_unfiltered_file, sep="\t", index=False)

        df_UMI_count_filter = df_UMI_count.groupby(
            ["barcode", "chain"], as_index=False
        ).head(1)
        # out filtered
        df_UMI_count_filter.to_csv(self.UMI_count_filtered_file, sep="\t", index=False)


def mapping_vdj(args):
    with Mapping_vdj(args, display_title="Mapping") as runner:
        runner.run()


def get_opts_mapping_vdj(parser, sub_program):
    parser.add_argument("--ref_path", help="reference path for igblast")
    parser.add_argument("--type", help="TCR or BCR", required=True)
    parser.add_argument(
        "--species",
        choices=["human", "mouse"],
        help="Default human. human or mouse.",
        default="human",
    )
    parser.add_argument(
        "--split_fasta",
        help="split fasta file to map to avoid running out of memory.",
        action="store_true",
    )
    if sub_program:
        parser.add_argument(
            "--fasta",
            help="Required. Input fasta file.",
            required=True,
        )
        parser = s_common(parser)



"""
Script to analyze human-mouse mixed single-cell RNA-seq data and identify doublets.
Generates a summary table with counts of human cells, mouse cells, doublets, and ambients.
Split the matrix into human and mouse based on gene names.
"""

import sys
from celescope.tools.matrix import CountMatrix
import pandas as pd
import argparse
import glob
import os
import functools


def run(mtx_path, doublet_threshold: float) -> dict:
    """
    Split human and mouse cells based on gene names and identify doublets.
    """
    mtx = CountMatrix.from_matrix_dir(mtx_path)
    sample = mtx_path.split("/")[-3]
    m = mtx.get_matrix()
    f = mtx.get_features()

    human_genes = []
    mouse_genes = []
    for i, gn in enumerate(f.gene_name):
        if gn == gn.upper():
            human_genes.append(i)
        else:
            mouse_genes.append(i)
    m_csr = m.tocsr()
    umi_h = m_csr[human_genes, :].sum(axis=0)
    umi_m = m_csr[mouse_genes, :].sum(axis=0)
    df_m = pd.DataFrame(umi_m.T, columns=["mouse"])
    df_h = pd.DataFrame(umi_h.T, columns=["human"])
    df = pd.merge(df_m, df_h, left_index=True, right_index=True)
    df["umi_sum"] = df.sum(axis=1)
    df["human_percent"] = df["human"] / df["umi_sum"] * 100
    df["mouse_percent"] = df["mouse"] / df["umi_sum"] * 100
    df["identity"] = "doublet"
    df.loc[df["human_percent"] < doublet_threshold, "identity"] = "mouse"
    df.loc[df["mouse_percent"] < doublet_threshold, "identity"] = "human"
    df.loc[df["identity"] == "human", "ambient_percent"] = df[
        df["identity"] == "human"
    ].mouse_percent
    df.loc[df["identity"] == "mouse", "ambient_percent"] = df[
        df["identity"] == "mouse"
    ].human_percent

    n_cell = df.shape[0]
    n_doublet = df[df["identity"] == "doublet"].shape[0]
    doublet_percent = round(n_doublet / n_cell * 100, 2)
    dict = {
        "n_cell": n_cell,
        "n_human_cell": df[df["identity"] == "human"].shape[0],
        "n_mouse_cell": df[df["identity"] == "mouse"].shape[0],
        "doublet_umi_percent_threshold": doublet_threshold,
        "n_doublet": n_doublet,
        "doublet_cell_percent": doublet_percent,
        "median_ambient_umi_percent": round(df["ambient_percent"].median(), 2),
        "mean_ambient_umi_percent": round(df["ambient_percent"].mean(), 2),
    }

    # split matrix
    human_barcodes = df[df["identity"] == "human"].index.tolist()
    mouse_barcodes = df[df["identity"] == "mouse"].index.tolist()
    out_dir = sample + ".out"
    mtx.slice_matrix(human_barcodes).to_matrix_dir(
        os.path.join(out_dir, "human_filtered")
    )
    mtx.slice_matrix(mouse_barcodes).to_matrix_dir(
        os.path.join(out_dir, "mouse_filtered")
    )

    return dict


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--celescope_dir",
        type=str,
        help="comma separated celescope output dirs",
        required=True,
    )
    parser.add_argument(
        "--doublet_threshold",
        type=float,
        default=25.0,
        help="threshold to define doublet cells",
    )
    args = parser.parse_args()

    cur_dir = os.getcwd()
    dfs = []
    matrix_path = set()
    for d in args.celescope_dir.split(","):
        paths = glob.glob(os.path.join(d, "*/*/filtered"))
        matrix_path.update(paths)
    if not matrix_path:
        sys.exit(
            "No filtered matrix found in the provided celescope directories. The filtered matrix should be located at <celescope_dir>/<sample>/outs/filtered."
        )
    print("Analyzing matrices from paths:", matrix_path)

    for path in matrix_path:
        sample = path.split("/")[-3]
        dict = run(path, args.doublet_threshold)
        dfs.append(pd.DataFrame.from_dict(dict, orient="index", columns=[sample]))

    df = functools.reduce(
        lambda left, right: pd.merge(left, right, left_index=True, right_index=True),
        dfs,
    )
    df = df.T
    df = df.astype(
        {
            "n_cell": int,
            "n_human_cell": int,
            "n_mouse_cell": int,
            "n_doublet": int,
        }
    )
    os.chdir(cur_dir)
    current_time = pd.Timestamp.now().strftime("%Y%m%dT%H%M%S")
    outprefix = f"human_mouse_summary-{current_time}"
    df.to_csv(f"{outprefix}.tsv", sep="\t")


if __name__ == "__main__":
    main()



"""Split fastq file according to barcode. Append UMI to read name with umi_separator semicolon ':'"""

import io
from celescope.tools.step import Step, s_common
from celescope.tools import utils
import celescope.tools.parse_chemistry as parse_chemistry
import pysam
import sys
from celescope.chemistry_dict import chemistry_dict
from celescope.__init__ import HELP_DICT
from celescope.bulk_rna.starsolo import get_barcode_sample


UMI_SEPARATOR = ":"


class Split(Step):
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.split_fastq = args.split_fastq
        self.split_bam = args.split_bam

        fq1_list = args.fq1.split(",")
        chemistry = parse_chemistry.get_chemistry(self.assay, args.chemistry, fq1_list)
        self.pattern_dict, bc = parse_chemistry.get_pattern_dict_and_bc(
            chemistry, args.pattern, args.whitelist
        )
        self.raw_list, self.mismatch_list = (
            parse_chemistry.create_mismatch_origin_dicts_from_whitelists(bc, 1)
        )

        self.barcode_sample = get_barcode_sample(bc[0], args.well_sample)

    @utils.add_log
    def run(self):
        fq_dict = {}
        bam_dict = {}
        all_bam = pysam.AlignmentFile(self.args.bam, "rb")
        for barcode, sample in self.barcode_sample.items():
            if self.split_fastq:
                file_obj = utils.generic_open(
                    f"{self.outdir}/{sample}.fq.gz", "wb", compresslevel=1
                )
                fq_dict[barcode] = io.BufferedWriter(
                    file_obj, buffer_size=16 * 1024 * 1024
                )
            if self.split_bam:
                bam_dict[barcode] = pysam.AlignmentFile(
                    f"{self.outdir}/{sample}.bam", "wb", header=all_bam.header
                )

        dup_align_read_names = set()
        for segment in all_bam:
            cb = segment.get_tag("CB")
            if cb == "-":
                continue
            if cb not in self.barcode_sample:
                continue
            if self.split_bam:
                bam_dict[cb].write(segment)
            if segment.get_tag("NH") > 1:
                if segment.query_name in dup_align_read_names:
                    continue
                else:
                    dup_align_read_names.add(segment.query_name)
            if self.split_fastq:
                umi = segment.get_tag("UB")
                name = segment.query_name + UMI_SEPARATOR + umi
                seq = segment.get_forward_sequence()
                qual = "".join(chr(q + 33) for q in segment.get_forward_qualities())
                fq_dict[cb].write(utils.fastq_line(name, seq, qual).encode())


@utils.add_log
def split(args):
    if not args.split_fastq:
        sys.stderr.write("--split_fastq not set. Skip split_fastq\n")
    if not args.split_bam:
        sys.stderr.write("--split_bam not set. Skip split_bam\n")
    if not (args.split_fastq or args.split_bam):
        return
    Split(args).run()


def get_opts_split(parser, sub_program):
    parser.add_argument(
        "--chemistry",
        help=HELP_DICT["chemistry"],
        choices=list(chemistry_dict.keys()),
        default="auto",
    )
    parser.add_argument(
        "--pattern",
        help="""The pattern of R1 reads, e.g. `C8L16C8L16C8L1U12T18`. The number after the letter represents the number 
        of bases.  
        - `C`: cell barcode  
        - `L`: linker(common sequences)  
        - `U`: UMI    
        - `T`: poly T""",
    )
    parser.add_argument(
        "--whitelist",
        help="Cell barcode whitelist file path, one cell barcode per line.",
    )
    parser.add_argument(
        "--well_sample",
        help="tsv file of well numbers and sample names. The first column is well numbers and the second column is sample names.",
        required=True,
    )
    parser.add_argument(
        "--split_fastq",
        help="Split fastq file according to well barcodes. Append UMI to read name with umi_separator semicolon ':'",
        action="store_true",
    )
    parser.add_argument(
        "--split_bam",
        help="Split BAM file according to well barcodes.",
        action="store_true",
    )
    if sub_program:
        parser.add_argument(
            "--fq1",
            help="R1 fastq file. Multiple files are separated by comma.",
            required=True,
        )
        parser.add_argument(
            "--bam",
            help="BAM file",
            required=True,
        )
        parser.add_argument(
            "--barcodes",
            help="barcode file.",
            required=True,
        )
        s_common(parser)
    return parser



from pathlib import Path
from celescope.tools import utils
from celescope.__init__ import __VERSION__
from celescope.chemistry_dict import chemistry_dict
import subprocess
import tarfile

from celescope.tools.step import Step, s_common

from celescope.tools.parse_chemistry import get_chemistry, invalid_debug


def sub_fq(fq, out_fq, use_read=1000, skip_read=10000):
    cmd = f"zcat {fq} | head -n {(use_read + skip_read) * 4} | tail -n {use_read * 4} | gzip > {out_fq}"
    subprocess.check_call(cmd, shell=True)


class Sample(Step):
    def __init__(self, args):
        Step.__init__(self, args)
        self.fq1_list = args.fq1.split(",")
        self.fq2_list = args.fq2.split(",")
        # out
        self.invalid_htmls: list[Path] = []
        self.subsample_fqs: list[Path] = []

    @utils.add_log
    def run(self):
        chemistry = get_chemistry(self.assay, self.args.chemistry, self.fq1_list)

        self.add_metric(
            name="Sample ID",
            value=self.sample,
        )
        self.add_metric(
            name="Assay",
            value=self.assay,
        )
        self.add_metric(
            name="Chemistry",
            value=chemistry,
            help_info='For more information, see <a href="https://github.com/singleron-RD/CeleScope/blob/master/doc/chemistry.md">here</a>',
        )
        self.add_metric(
            name="Software Version",
            value=__VERSION__,
        )

        self.subsample(self.fq1_list)
        self.subsample(self.fq2_list)
        self.debug_invalid(chemistry)
        self.pack_files()

    def subsample(self, fq_list):
        for fq in fq_list:
            fq = Path(fq)
            out_fq = Path(self.outdir) / f"{fq.stem}.subsample.gz"
            sub_fq(fq, out_fq, self.args.use_read, self.args.skip_read)
            self.subsample_fqs.append(out_fq)

    def pack_files(self):
        with tarfile.open(f"{self.out_prefix}_debug.tar", "w") as tar:
            for file in self.subsample_fqs + self.invalid_htmls:
                tar.add(file, arcname=file.name)

    def debug_invalid(self, chemistry):
        for fq1 in self.fq1_list:
            fq1 = Path(fq1)
            out_html = Path(self.outdir) / f"{fq1.stem}.invalid.html"
            invalid_debug(
                chemistry,
                fq1,
                out_html,
                use_read=self.args.use_read,
                skip_read=self.args.skip_read,
                pattern=self.args.pattern,
                whitelist=self.args.whitelist,
                linker=self.args.linker,
            )
            self.invalid_htmls.append(out_html)


@utils.add_log
def sample(args):
    with Sample(args) as runner:
        runner.run()


def get_opts_sample(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument("--fq1", help="read1 fq file")
        parser.add_argument("--fq2", help="read2 fq file")
        parser.add_argument("--linker")
        parser.add_argument(
            "--use_read", help="number of reads used for debug", default=1000, type=int
        )
        parser.add_argument(
            "--skip_read",
            help="Number of reads to skip for debugging. The R1 FASTQ contains N bases in the first hundred reads, so it is better to skip them during debugging",
            default=10000,
            type=int,
        )
    parser.add_argument(
        "--chemistry",
        choices=list(chemistry_dict.keys()),
        help="chemistry version",
        default="auto",
    )
    parser.add_argument("--pattern")
    parser.add_argument(
        "--whitelist",
    )
    return parser
